import logging

from shraga_common import ShragaConfig

logger = logging.getLogger(__name__)

SHRAGA_CONFIG = None

def load_config(path):
    global SHRAGA_CONFIG
    SHRAGA_CONFIG = ShragaConfig().load(path)
    return SHRAGA_CONFIG

def get_config(k: str = None, default=None):
    if not SHRAGA_CONFIG:
        logger.error("Shraga config not loaded")
        return default
    if not k:
        return SHRAGA_CONFIG
    return SHRAGA_CONFIG.get(k, default)



from fastapi import APIRouter

from shraga_common.embedders import BedrockEmbedder
from shraga_common.retrievers import (ElasticsearchRetriever,
                                      OpenSearchRetriever)

from ..config import get_config


router = APIRouter()


@router.get("/")
async def list_services() -> dict:
    shraga_config = get_config()
    d = dict()
    try:
        embedder = BedrockEmbedder(shraga_config)
        d["embedder"] = "bedrock"
    except Exception:
        pass

    for retriever_name, r in shraga_config.retrievers().items():
        try:
            if r.get("type") == "opensearch":
                retriever = OpenSearchRetriever(shraga_config)
            elif r.get("type") == "elasticsearch":
                retriever = ElasticsearchRetriever(shraga_config)
            else:
                continue

            d["retriever:" + retriever_name] = {
                "type": r.get("type"),
                "test": await retriever.execute_empty_query(),
            }
        except Exception as e:
            d["retriever:" + retriever_name] = {
                "type": r.get("type"),
                "failure": str(e),
            }

    return d



import os
import re
import sys

import yaml
from pydash import _


class ShragaConfig:
    path_matcher = re.compile(r"\$\{([^}^{]+)}")

    def __init__(self):
        self.all_configs = None

    def path_constructor(self, loader, node):
        """Extract the matched value, expand env variable, and replace the match"""
        value = node.value
        match = self.path_matcher.match(value)
        env_var = match.group()[2:-1]
        env_var_val = os.environ.get(env_var)
        if env_var_val:
            return env_var_val + value[match.end() :]
        return ""
    
    def validate_config(self):
        """Validate configuration according to business rules"""
        list_flows = self.get("ui.list_flows")
        default_flow = self.get("ui.default_flow")
        
        # If list_flows is not True (either False or None), default_flow must be set
        if list_flows is not True:
            if not default_flow:
                raise ValueError("When list_flows is not enabled, default_flow must be specified")

    def load(self, config_path: str = None):
        if not config_path:
            config_path = os.getenv("CONFIG_PATH") or "config.yaml"
        print(f"Loading config from {config_path}")

        try:
            with open(config_path, encoding='utf-8') as stream:
                yaml.add_implicit_resolver("!path", self.path_matcher)
                yaml.add_constructor("!path", self.path_constructor)
                self.all_configs = yaml.load(stream, Loader=yaml.FullLoader)

                self.validate_config()

                return self
            
        except ValueError as e:
            print(f"ERROR: Configuration validation failed: {e}", file=sys.stderr)
            return None
        except (FileNotFoundError, yaml.YAMLError, IOError) as e:
            print(f"ERROR: Failed to load config from {config_path}: {e}", file=sys.stderr)
            return None

    def get(self, k: str, default=None):
        v = _.get(self.all_configs, k)
        if v is None:
            return default
        return v

    def set(self, k: str, v):
        _.set(self.all_configs, k, v)

    def auth_realms(self):
        return _.get(self.all_configs, "auth.realms") or dict()

    def auth_users(self):
        return set(_.get(self.all_configs, "auth.users") or [])

    def retrievers(self) -> dict:
        return _.get(self.all_configs, "retrievers") or dict()



import json
import logging
import os
import uuid
from typing import Any, List, Optional

from openai import OpenAI, OpenAIError

from shraga_common import ShragaConfig

from .base_embedder import BaseEmbedder, BaseEmbedderGenerateOptions

# Configure logging
logger = logging.getLogger(__name__)

OPENAI_EMBEDDING_MODEL_IDS = ["text-embedding-3-small", "text-embedding-3-large"]


class OpenAIEmbedder(BaseEmbedder):
    def __init__(
        self,
        shraga_config: Optional[ShragaConfig] = None,
        openai_api_key: Optional[str] = None,
    ):
        openai_api_key = openai_api_key or shraga_config.get("services.openai.api_key")
        self.client = OpenAI(api_key=openai_api_key)

    async def generate_vector(
        self, text: str, extra_options: Optional[BaseEmbedderGenerateOptions] = None
    ) -> Any:
        if not extra_options:
            extra_options = {}

        model_name = extra_options.get("embed_model_name", "text-embedding-3-small")

        if model_name not in OPENAI_EMBEDDING_MODEL_IDS:
            raise ValueError(
                f"Invalid model ID. Available model IDs are: {', '.join(OPENAI_EMBEDDING_MODEL_IDS)}"
            )

        try:
            return (
                self.client.embeddings.create(input=[text], model=model_name)
                .data[0]
                .embedding
            )
        except OpenAIError as e:
            raise RuntimeError(f"OpenAI API returned an error: {e}")
        except Exception as e:
            raise RuntimeError(f"An unexpected error occurred: {e}")

    def generate_vectors_batch(self, requests: List[dict]) -> Any:
        try:
            filename = f"batch_{uuid.uuid4().hex}.jsonl"
            with open(filename, "w") as f:
                for request in requests:
                    f.write(json.dumps(request) + "\n")

            input_file = self.client.files.create(
                file=open(filename, "rb"), purpose="batch"
            )

            os.remove(filename)

            batch_response = self.client.batches.create(
                input_file_id=input_file.id,
                endpoint="/v1/embeddings",
                completion_window="24h",
            )

            return batch_response

        except OpenAIError as e:
            logger.error(f"OpenAI API returned an error: {e}")
            raise RuntimeError(f"OpenAI API error: {e}")

    def poll_batch_job(self, job_id: str) -> Any:
        try:
            job_status = self.client.batches.retrieve(job_id)
            return job_status
        except OpenAIError as e:
            logger.error(f"Error retrieving batch job {job_id}: {e}")
            raise RuntimeError(f"Error retrieving batch job {job_id}: {e}")

    def cancel_batch_job(self, batch_id: str) -> None:
        try:
            self.client.batches.cancel(batch_id)
            logger.info(f"Batch {batch_id} canceled successfully.")
        except OpenAIError as e:
            logger.error(f"Error canceling batch {batch_id}: {e}")
            raise RuntimeError(f"Error canceling batch {batch_id}: {e}")

    def delete_file(self, file_id: str) -> None:
        try:
            deleted = self.client.files.delete(file_id)
            deleted_msg = "deleted successfully" if deleted.deleted else "not deleted"
            logger.info(f"File {file_id} {deleted_msg}.")
        except OpenAIError as e:
            logger.error(f"Error deleting file {file_id}: {e}")
            raise RuntimeError(f"Error deleting file {file_id}: {e}")

    def get_output_file_content(self, output_file_id: str) -> List[str]:
        try:
            return [
                json.loads(line)
                for line in self.client.files.content(output_file_id).text.splitlines()
            ]
        except OpenAIError as e:
            logger.error(f"Error retrieving output file {output_file_id}: {e}")
            raise RuntimeError(f"Error retrieving output file {output_file_id}: {e}")

    def list_batches(self) -> List[dict]:
        try:
            return self.client.batches.list()
        except OpenAIError as e:
            logger.error(f"Error listing batches: {e}")
            raise RuntimeError(f"Error listing batches: {e}")



from abc import abstractmethod
from typing import Literal, Optional, TypedDict

LlmModelProvider = Literal["cohere", "openai"]


class LLMServiceOptions(TypedDict):
    model_id: str
    system_prompt: Optional[str]


class LLMService:
    @abstractmethod
    async def invoke_model(
        self, prompt: str, options: Optional[LLMServiceOptions] = None
    ) -> str:
        pass



from datetime import date
from typing import Dict, List, Literal, Optional

from pydantic import BaseModel

from shraga_common.services import LlmModelProvider


class EvaluationModel(BaseModel):
    input_files: List[str]
    evaluating_model: Literal[LlmModelProvider] = "cohere"
    flow_id: str
    output_cluster: Optional[str] = None
    test_count: Optional[int] = 1
    max_concurrent_tests: Optional[int] = 4
    last_lookup_date: Optional[date] = None
    tag: Optional[str] = None
    run_only: bool = False
    evaluated_flow_preferences: Optional[Dict] = None


class EvaluationScenario(BaseModel):
    question: str
    answer: Optional[str] = None
    metadata: Optional[Dict] = None



from .base_embedder import BaseEmbedder
from .bedrock_embedder import BedrockEmbedder
from .cohere_embedder import CohereEmbedder

# commented out due to slow loading!
# from .google_embedder import GoogleEmbedder
from .openai_embedder import OpenAIEmbedder
from .types import EmbedderModelName, EmbedderModelProvider

__all__ = [
    "BaseEmbedder",
    "EmbedderModelName",
    "EmbedderModelProvider",
    "BedrockEmbedder",
    "OpenAIEmbedder",
    "CohereEmbedder",
]



import unittest
from unittest.mock import Mock, patch
import sys
import os

# Add the project root to the path to help with imports
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../..'))

from shraga_common.app.auth.user import ShragaUser
from shraga_common.app.services.history_service import log_interaction, log_system_message

class TestLogInteraction(unittest.IsolatedAsyncioTestCase):

    def create_mock_request(self, user_id: str):
        request = Mock()
        if user_id != "<unknown>":
            # Create a ShragaUser instead of a basic Mock
            request.user = ShragaUser(
                username=user_id,
                roles=["user"],
                metadata={"auth_type": "test"}
            )
        else:
            # For the case where we test without a user
            pass
        request.headers = {"user-agent": "test-agent"}
        return request

    @patch('shraga_common.app.services.history_service.get_config')
    @patch('shraga_common.app.services.history_service.get_history_client')
    @patch('shraga_common.logger.get_config_info')
    @patch('shraga_common.logger.get_platform_info')
    @patch('shraga_common.logger.get_user_agent_info')
    async def test_user_org_added_to_log_document(
        self, 
        mock_get_user_agent_info,
        mock_get_platform_info,
        mock_get_config_info,
        mock_get_history_client, 
        mock_get_config
    ):
        test_cases = [
            ("alice@techcorp.com", "techcorp.com"),
            ("user@gmail.com", ""),
            ("username123", ""),
        ]
        
        for user_id, expected_org in test_cases:
            with self.subTest(user_id=user_id, expected_org=expected_org):
                mock_opensearch_client = Mock()
                mock_get_history_client.return_value = (mock_opensearch_client, "test_index")
                mock_get_config.return_value = {"test": "config"}
                mock_get_config_info.return_value = {"config": "test"}
                mock_get_platform_info.return_value = {"platform": "test"}
                mock_get_user_agent_info.return_value = {"user_agent": "test"}
                
                request = self.create_mock_request(user_id)
                context = {"text": "test message", "chat_id": "test_chat", "flow_id": "test_flow"}
                
                result = await log_interaction("user", request, context)
                
                self.assertTrue(result)
                
                mock_opensearch_client.index.assert_called_once()
                
                call_args = mock_opensearch_client.index.call_args
                self.assertEqual(call_args[1]["index"], "test_index")
                
                saved_document = call_args[1]["body"]
                self.assertEqual(saved_document["user_org"], expected_org)
                self.assertEqual(saved_document["text"], "test message")

    @patch('shraga_common.app.services.history_service.get_config')
    @patch('shraga_common.app.services.history_service.get_history_client')
    @patch('shraga_common.logger.get_config_info')
    @patch('shraga_common.logger.get_platform_info')
    @patch('shraga_common.logger.get_user_agent_info')
    async def test_handles_request_without_user(
        self,
        mock_get_user_agent_info,
        mock_get_platform_info,
        mock_get_config_info,
        mock_get_history_client,
        mock_get_config
    ):
        mock_opensearch_client = Mock()
        mock_get_history_client.return_value = (mock_opensearch_client, "test_index")
        mock_get_config.return_value = {"test": "config"}
        mock_get_config_info.return_value = {"config": "test"}
        mock_get_platform_info.return_value = {"platform": "test"}
        mock_get_user_agent_info.return_value = {"user_agent": "test"}
        
        # Creating a request without a user attribute
        request = Mock(spec=['headers'])
        request.headers = {"user-agent": "test-agent"}
        # Make request.user raise AttributeError when accessed
        def user_property_raiser(obj):
            raise AttributeError("'Request' object has no attribute 'user'")
        
        # Create a property that raises an AttributeError when accessed
        type(request).__getattr__ = Mock(side_effect=user_property_raiser)
        
        context = {"text": "test", "chat_id": "test_chat", "flow_id": "test_flow"}
        with patch('shraga_common.app.services.history_service.ShragaUser') as mock_shraga_user:
            # Create a mock ShragaUser with default values
            mock_anonymous_user = Mock()
            mock_anonymous_user.identity = "<unknown>"
            mock_anonymous_user.user_org = ""
            mock_anonymous_user.metadata = {}
            mock_shraga_user.return_value = mock_anonymous_user
            
            result = await log_interaction("user", request, context)
            
            self.assertTrue(result)
            mock_shraga_user.assert_called_once()
            
            saved_document = mock_opensearch_client.index.call_args[1]["body"]
            self.assertEqual(saved_document["user_id"], "<unknown>")
            self.assertEqual(saved_document["user_org"], "")

    @patch('shraga_common.app.services.history_service.get_config')
    @patch('shraga_common.app.services.history_service.get_history_client')
    @patch('shraga_common.logger.get_config_info')
    @patch('shraga_common.logger.get_platform_info')
    @patch('shraga_common.logger.get_user_agent_info')
    async def test_log_interaction_no_client_returns_false(
        self,
        mock_get_user_agent_info,
        mock_get_platform_info,
        mock_get_config_info,
        mock_get_history_client,
        mock_get_config
    ):
        """Test that log_interaction returns False when no history client is available."""
        # Simulate no history client available
        mock_get_history_client.return_value = (None, None)
        
        request = self.create_mock_request("test@example.com")
        context = {"text": "test message", "chat_id": "test_chat"}
        
        result = await log_interaction("user", request, context)
        
        self.assertFalse(result)
        # Ensure no indexing was attempted
        mock_get_config_info.assert_not_called()
        mock_get_platform_info.assert_not_called()

    @patch('shraga_common.app.services.history_service.get_config')
    @patch('shraga_common.app.services.history_service.get_history_client')
    @patch('shraga_common.logger.get_config_info')
    @patch('shraga_common.logger.get_platform_info')
    @patch('shraga_common.logger.get_user_agent_info')
    async def test_log_interaction_handles_indexing_exception(
        self,
        mock_get_user_agent_info,
        mock_get_platform_info,
        mock_get_config_info,
        mock_get_history_client,
        mock_get_config
    ):
        """Test that log_interaction handles exceptions during indexing gracefully."""
        mock_opensearch_client = Mock()
        mock_opensearch_client.index.side_effect = Exception("Indexing failed")
        mock_get_history_client.return_value = (mock_opensearch_client, "test_index")
        mock_get_config.return_value = {"test": "config"}
        mock_get_config_info.return_value = {"config": "test"}
        mock_get_platform_info.return_value = {"platform": "test"}
        mock_get_user_agent_info.return_value = {"user_agent": "test"}
        
        request = self.create_mock_request("test@example.com")
        context = {"text": "test message", "chat_id": "test_chat"}
        
        with patch('shraga_common.app.services.history_service.logger') as mock_logger:
            result = await log_interaction("user", request, context)
            
            self.assertFalse(result)
            mock_logger.exception.assert_called_once()

    @patch('shraga_common.app.services.history_service.get_config')
    @patch('shraga_common.app.services.history_service.log_interaction')
    async def test_log_system_message_history_disabled(
        self,
        mock_log_interaction,
        mock_get_config
    ):
        """Test that log_system_message returns False when history is disabled."""
        mock_get_config.return_value = False  # history.enabled = False
        
        request = self.create_mock_request("test@example.com")
        request_body = Mock()
        request_body.msg_id = "msg_123"
        request_body.chat_id = "chat_123"
        request_body.flow_id = "flow_123"
        request_body.position = 1
        request_body.preferences = {}
        
        response = Mock()
        response.response_text = "System response"
        response.stats = []
        response.payload = {}
        response.retrieval_results = []
        response.trace = []
        
        result = await log_system_message(request, request_body, response)
        
        self.assertFalse(result)
        mock_log_interaction.assert_not_called()

    @patch('shraga_common.app.services.history_service.get_config')
    @patch('shraga_common.app.services.history_service.log_interaction')
    async def test_log_system_message_cleans_retrieval_results(
        self,
        mock_log_interaction,
        mock_get_config
    ):
        """Test that log_system_message cleans extra data from retrieval results."""
        mock_get_config.return_value = True  # history.enabled = True
        mock_log_interaction.return_value = True
        
        request = self.create_mock_request("test@example.com")
        request_body = Mock()
        request_body.msg_id = "msg_123"
        request_body.chat_id = "chat_123"
        request_body.flow_id = "flow_123"
        request_body.position = 1
        request_body.preferences = {"theme": "dark"}
        
        # Create mock retrieval results with extra data
        retrieval_result1 = Mock()
        retrieval_result1.extra = {"sensitive": "data"}
        retrieval_result2 = Mock()
        retrieval_result2.extra = {"more": "sensitive_data"}
        
        response = Mock()
        response.response_text = "System response"
        response.stats = [{"metric": "value"}]
        response.payload = {"key": "value"}
        response.retrieval_results = [retrieval_result1, retrieval_result2]
        response.trace = ["trace1", "trace2"]
        
        result = await log_system_message(request, request_body, response)
        
        self.assertTrue(result)
        
        # Verify extra data was cleaned
        self.assertEqual(retrieval_result1.extra, {})
        self.assertEqual(retrieval_result2.extra, {})
        
        # Verify log_interaction was called with correct parameters
        mock_log_interaction.assert_called_once_with(
            "system",
            request,
            {
                "msg_id": "msg_123",
                "chat_id": "chat_123",
                "flow_id": "flow_123",
                "text": "System response",
                "position": 2,  # position + 1
                "preferences": {"theme": "dark"},
                "stats": [{"metric": "value"}],
                "payload": {"key": "value"},
                "retrieval_results": [retrieval_result1, retrieval_result2],
                "trace": ["trace1", "trace2"],
            },
        )

    @patch('shraga_common.app.services.history_service.get_config')
    @patch('shraga_common.app.services.history_service.log_interaction')
    async def test_log_system_message_handles_none_response(
        self,
        mock_log_interaction,
        mock_get_config
    ):
        """Test that log_system_message handles None response gracefully."""
        mock_get_config.return_value = True  # history.enabled = True
        mock_log_interaction.return_value = True
        
        request = self.create_mock_request("test@example.com")
        request_body = Mock()
        request_body.msg_id = "msg_123"
        request_body.chat_id = "chat_123"
        request_body.flow_id = "flow_123"
        request_body.position = 5
        request_body.preferences = {}
        
        result = await log_system_message(request, request_body, None)
        
        self.assertTrue(result)
        
        # Verify log_interaction was called with default values for None response
        mock_log_interaction.assert_called_once_with(
            "system",
            request,
            {
                "msg_id": "msg_123",
                "chat_id": "chat_123",
                "flow_id": "flow_123",
                "text": "",
                "position": 6,  # position + 1
                "preferences": {},
                "stats": None,
                "payload": None,
                "retrieval_results": None,
                "trace": None,
            },
        )

    @patch('shraga_common.app.services.history_service.get_config')
    @patch('shraga_common.app.services.history_service.log_interaction')
    async def test_log_system_message_handles_none_position(
        self,
        mock_log_interaction,
        mock_get_config
    ):
        """Test that log_system_message handles None position in request_body."""
        mock_get_config.return_value = True  # history.enabled = True
        mock_log_interaction.return_value = True
        
        request = self.create_mock_request("test@example.com")
        request_body = Mock()
        request_body.msg_id = "msg_123"
        request_body.chat_id = "chat_123"
        request_body.flow_id = "flow_123"
        request_body.position = None  # None position
        request_body.preferences = {}
        
        response = Mock()
        response.response_text = "System response"
        response.stats = []
        response.payload = {}
        response.retrieval_results = None
        response.trace = []
        
        result = await log_system_message(request, request_body, response)
        
        self.assertTrue(result)
        
        # Verify position calculation handles None
        mock_log_interaction.assert_called_once()
        call_args = mock_log_interaction.call_args[0][2]  # Get the context dict
        self.assertEqual(call_args["position"], 1)  # (None or 0) + 1 = 1


if __name__ == '__main__':
    unittest.main()


import logging
import traceback
from datetime import datetime
from nanoid import generate as nanoid

from fastapi import APIRouter, BackgroundTasks, HTTPException, Request
from fastapi.responses import JSONResponse

from shraga_common.models import FlowBase, FlowResponse

from ..config import get_config
from shraga_common import RequestCancelledException, LLMServiceUnavailableException
from ..models import FlowRunApiRequest
from .request_utils import execute_cancellable_flow
from ..services import history_service, list_flows_service
from ..utils import clean_input

logger = logging.getLogger(__name__)
router = APIRouter()

def add_log_error_task(bg_tasks: BackgroundTasks, keep: bool, request: Request, req_body: FlowRunApiRequest, text: str):
    if keep:
        bg_tasks.add_task(
            history_service.log_interaction,
            "error",
            request,
            {
                "chat_id": req_body.chat_id,
                "flow_id": req_body.flow_id,
                "text": text,
                "traceback": traceback.format_exc()
            }
        )

async def run_flow(
    request: Request,
    req_body: FlowRunApiRequest,
    bg_tasks: BackgroundTasks,
    keep: bool = True,
) -> FlowResponse:
    max_length = get_config("flows.input_max_length", 1000)
    if len(req_body.question) > max_length:
        return JSONResponse(
            status_code=400,
            content={
                "detail": f"Message exceeds maximum length of {max_length} characters"
            },
        )

    req_body.question = clean_input(req_body.question)
    req_body.msg_id = nanoid(size=12)
    
    if not req_body.chat_id:
        req_body.chat_id = f"chat_{datetime.now().strftime('%Y%m%d%H%M%S')}"
    
    available_flows = list_flows_service.get_available_flows()

    f = available_flows.get(req_body.flow_id)
    if not f:
        raise HTTPException(
            status_code=400, detail="Unknown flow id: " + req_body.flow_id
        )
        
    shraga_config = get_config()
    flow: FlowBase = f(shraga_config, flows=available_flows)

    if keep:
        await history_service.log_user_message(request, req_body)

    try:
        rsp = await execute_cancellable_flow(request, flow, req_body)

        if hasattr(req_body, "chat_id"):
            rsp.chat_id = req_body.chat_id

        if keep:
            if rsp.stats:
                # log flow execution
                for stat in rsp.stats:
                    bg_tasks.add_task(history_service.log_flow, request, req_body, stat)
            # we need to run this synchronously to avoid a race condition with the FE fetching the history
            await history_service.log_system_message(request, req_body, response=rsp)

        return rsp

    except RequestCancelledException:
        return JSONResponse(status_code=500, content={"detail": "Request cancelled by client"})
    
    except LLMServiceUnavailableException as e:
        add_log_error_task(bg_tasks, keep, request, req_body, str(e))
        return JSONResponse(status_code=503, content={"detail": str(e)})

    except Exception as e:
        add_log_error_task(bg_tasks, keep, request, req_body, str(e))
        logger.exception("Error running flow:", exc_info=e)
        # Important: do not raise an exception here, otherwise background tasks will not be executed
        return JSONResponse(status_code=500, content={"detail": str(e)})



from .exceptions import ShragaException
from .logging import init_logging
from .utils import (get_config_info, get_git_branch, get_git_commit,
                    get_platform_info, get_user_agent_info)

__all__ = [
    "get_platform_info",
    "get_config_info",
    "get_user_agent_info",
    "get_git_commit",
    "get_git_branch",
    "ShragaException",
    "init_logging",
]



from typing import Any, Dict, Optional

from starlette.authentication import SimpleUser

from shraga_common.utils import extract_user_org


class ShragaUser(SimpleUser):
    """
    ShragaUser extends SimpleUser to provide additional functionality
    for Shraga application users.
    
    This class adds support for storing additional user information,
    such as roles, and any user-specific metadata.
    """
    
    def __init__(
        self,
        username: str,
        roles: Optional[list] = None,
        metadata: Optional[Dict[str, Any]] = None):
        super().__init__(username)
        self.user_org = extract_user_org(self.username)
        self.roles = roles or []
        self.metadata = metadata or {}
    
    @property
    def identity(self) -> str:
        return self.username
    
    def has_role(self, role: str) -> bool:
        return role in self.roles
    
    def get_metadata(self, key: str, default: Any = None) -> Any:
        """
        Get a metadata value by key.
        
        Args:
            key: The metadata key
            default: The default value to return if the key is not found
            
        Returns:
            The metadata value, or the default if not found
        """
        return self.metadata.get(key, default)
    
    def set_metadata(self, key: str, value: Any) -> None:
        """
        Set a metadata value.
        
        Args:
            key: The metadata key
            value: The value to set
        """
        self.metadata[key] = value
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ShragaUser":
        return cls(
            username=data.get("username"),
            roles=data.get("roles"),
            metadata=data.get("metadata"),
        )
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "username": self.username,
            "roles": self.roles,
            "metadata": self.metadata,
        }
    
    def __str__(self) -> str:
        """String representation of the user."""
        return f"{self.username}"



import datetime
import logging
from typing import List, Optional

from fastapi import Request
from pydash import _

from shraga_common.logger import (get_config_info, get_platform_info,
                                   get_user_agent_info)
from shraga_common.models import FlowResponse, FlowStats
from shraga_common.utils import is_prod_env

from ..auth.user import ShragaUser
from ..config import get_config
from ..models import Chat, ChatMessage, FeedbackRequest, FlowRunApiRequest
from .get_history_client import get_history_client

logger = logging.getLogger(__name__)

# Constants
DEFAULT_CHAT_LIST_LENGTH = 50
DEFAULT_CHAT_MESSAGES_COUNT = 1000


# Helper functions
def _get_history_client_or_none():
    """Get history client and index, return None if not available."""
    try:
        shraga_config = get_config()
        client, index = get_history_client(shraga_config)
        return client, index if client else (None, None)
    except Exception as e:
        logger.exception("Error getting history client", exc_info=e)
        return None, None


def _get_user_from_request(request: Request) -> ShragaUser:
    """Extract user from request, return anonymous user if not available."""
    try:
        return request.user
    except (AttributeError, Exception):
        return ShragaUser(username="<unknown>")


def _build_chat_list_filters(user_id: str, start: Optional[str], end: Optional[str]) -> List[dict]:
    """Build filters for chat list query."""
    filters = []
    
    if user_id:
        filters.append({"term": {"user_id": user_id}})
        
    if start:
        filters.append({"range": {"timestamp": {"gte": start, "lte": end or "now"}}})

    if is_prod_env() and not user_id:
        filters.append({"term": {"config.prod": True}})
    
    return filters


def _build_chat_list_query(filters: List[dict]) -> dict:
    """Build Elasticsearch query for chat list."""
    return {
        "query": {
            "bool": {
                "must": [{"terms": {"msg_type": ["system", "user"]}}],
                "filter": filters,
            }
        },
        "size": 0,
        "aggs": {
            "by_chat": {
                "terms": {
                    "field": "chat_id",
                    "size": DEFAULT_CHAT_LIST_LENGTH,
                    "order": {"last_message": "desc"},
                },
                "aggs": {
                    "last_message": {"max": {"field": "timestamp"}},
                    "first_message": {"min": {"field": "timestamp"}},
                    "first": {
                        "top_hits": {
                            "size": 1,
                            "sort": [{"timestamp": {"order": "asc"}}]
                        }
                    },
                    "latest": {
                        "top_hits": {
                            "size": 1,
                            "sort": [{"timestamp": {"order": "desc"}}]
                        }
                    }
                }
            }
        }
    }


def _build_chat_messages_query(chat_id: str, count: int) -> dict:
    """Build Elasticsearch query for chat messages."""
    return {
        "query": {
            "bool": {
                "must": [
                    {"term": {"chat_id": chat_id}},
                    {"terms": {"msg_type": ["user", "system"]}}
                ]
            }
        },
        "sort": [{"timestamp": {"order": "desc"}}],
        "size": count
    }


# Main service functions
async def get_chat_list(
    user_id: str, 
    start: Optional[str] = None, 
    end: Optional[str] = None
) -> List[Chat]:
    """Get list of chats for a user with optional date filtering."""
    try:
        client, index = _get_history_client_or_none()
        if not client:
            return []
        
        filters = _build_chat_list_filters(user_id, start, end)
        query = _build_chat_list_query(filters)

        response = client.search(index=index, body=query)
        hits = _.get(response, "aggregations.by_chat.buckets") or []
        
        return [Chat.from_hit(hit) for hit in hits]

    except Exception as e:
        logger.exception("Error retrieving chat list for user %s", user_id, exc_info=e)
        return []
        

async def get_chat_messages(chat_id: str, count: int = DEFAULT_CHAT_MESSAGES_COUNT) -> List[ChatMessage]:
    """Get messages for a specific chat."""
    try:
        client, index = _get_history_client_or_none()
        if not client:
            return []
            
        query = _build_chat_messages_query(chat_id, count)
        response = client.search(index=index, body=query)
        
        hits = response.get("hits", {}).get("hits", [])
        messages = [ChatMessage.from_hit(hit) for hit in hits]
        messages.reverse()  # Reverse to get ascending order (oldest first)
        
        return messages
    
    except Exception as e:
        logger.exception("Error retrieving chat messages for chat %s", chat_id, exc_info=e)
        return []
        

async def get_chat(chat_id: str) -> Optional[Chat]:
    """Get a specific chat by ID."""
    try:
        client, index = _get_history_client_or_none()
        if not client:
            return None

        response = client.get(index=index, id=chat_id)
        if not response["found"]:
            return None
            
        return Chat(**response["_source"])
        
    except Exception as e:
        logger.exception("Error retrieving chat %s", chat_id, exc_info=e)
        return None


async def delete_chat(chat_id: str) -> bool:
    """Delete a chat by ID."""
    try:
        client, index = _get_history_client_or_none()
        if not client:
            return True  # Return True if history is disabled

        client.delete(index=index, id=chat_id)
        return True
        
    except Exception as e:
        logger.exception("Error deleting chat %s", chat_id, exc_info=e)
        return False


# Logging functions
async def log_interaction(msg_type: str, request: Request, context: dict) -> bool:
    """Log an interaction to the history system."""
    try:
        client, index = _get_history_client_or_none()
        if not client:
            return False

        user = _get_user_from_request(request)
        shraga_config = get_config()

        message = ChatMessage(
            msg_type=msg_type,
            timestamp=datetime.datetime.now(tz=datetime.timezone.utc),
            user_id=user.identity,
            **context
        )

        # Build the document to index
        doc = message.model_dump()
        doc.update({
            "platform": get_platform_info(),
            "config": get_config_info(shraga_config),
            "user_agent": get_user_agent_info(request.headers.get("user-agent")),
            "user_org": user.user_org,
            "user_metadata": user.metadata
        })

        client.index(index=index, body=doc)
        return True

    except Exception as e:
        logger.exception("Error logging interaction %s", msg_type, exc_info=e)
        return False


async def log_feedback(request: Request, request_body: FeedbackRequest) -> bool:
    """Log user feedback."""
    return await log_interaction(
        "feedback",
        request,
        {
            "msg_id": request_body.msg_id,
            "chat_id": request_body.chat_id,
            "flow_id": request_body.flow_id,
            "text": request_body.feedback_text,
            "position": request_body.position,
            "feedback": request_body.feedback,
        },
    )


async def log_user_message(request: Request, request_body: FlowRunApiRequest) -> bool:
    """Log user message."""
    return await log_interaction(
        "user",
        request,
        {
            "msg_id": request_body.msg_id,
            "chat_id": request_body.chat_id,
            "flow_id": request_body.flow_id,
            "text": request_body.question,
            "position": request_body.position,
            "preferences": request_body.preferences,
        },
    )


async def log_system_message(
    request: Request,
    request_body: FlowRunApiRequest,
    response: Optional[FlowResponse] = None,
) -> bool:
    """Log system response message."""
    history_enabled = get_config('history.enabled', False)
    if not history_enabled:
        return False
        
    # Clean up retrieval results before storing
    if response and response.retrieval_results:
        for result in response.retrieval_results:
            if result.extra:
                result.extra = {}
                
    return await log_interaction(
        "system",
        request,
        {
            "msg_id": request_body.msg_id,
            "chat_id": request_body.chat_id,
            "flow_id": request_body.flow_id,
            "text": response.response_text if response else "",
            "position": (request_body.position or 0) + 1,
            "preferences": request_body.preferences,
            "stats": response.stats if response else None,
            "payload": response.payload if response else None,
            "retrieval_results": response.retrieval_results if response else None,
            "trace": response.trace if response else None,
        },
    )


async def log_flow(request: Request, request_body: FlowRunApiRequest, stat: FlowStats) -> bool:
    """Log flow statistics."""
    return await log_interaction(
        "flow_stats",
        request,
        {
            "msg_id": request_body.msg_id,
            "chat_id": request_body.chat_id,
            "text": request_body.question,
            "flow_id": stat.flow_id,
            "stats": stat,
        },
    )


async def log_error_message(
    request: Request, 
    request_body: FlowRunApiRequest, 
    error: Exception, 
    traceback: str
) -> bool:
    """Log error message."""
    return await log_interaction(
        "error",
        request,
        {
            "msg_id": request_body.msg_id,
            "chat_id": request_body.chat_id,
            "flow_id": request_body.flow_id,
            "text": str(error),
            "traceback": traceback,
        },
    )



import argparse


def load_args():
    parser = argparse.ArgumentParser(
        description="Run the Preprocessing Preparation Flow"
    )
    parser.add_argument(
        "--embed",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Should the flow embed the text",
    )
    parser.add_argument(
        "--output_chunk_text",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Should the flow output the chunk text",
    )

    parser.add_argument(
        "--parallelism",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Should the flow run in parallel",
    )
    parser.add_argument(
        "--max_queue_size",
        type=int,
        default=50_000,
        help="The maximum queue size, if parallelism is enabled",
    )
    parser.add_argument(
        "--number_of_processes",
        type=int,
        default=20,
        help="The number of processes to run in parallel",
    )
    parser.add_argument(
        "--doc_type",
        type=str,
        help="Document type to process",
    )
    parser.add_argument(
        "--input_dir",
        required=True,
        type=str,
        help="Input dir",
    )
    parser.add_argument(
        "--input_file",
        type=str,
        help="Input file path",
    )
    parser.add_argument(
        "--skip_count",
        type=int,
        help="How many input files to skip",
    )
    args = parser.parse_args()

    print(args)
    return args



from typing import Optional, Literal
from pydantic import BaseModel
class RetrieverConfig(BaseModel):
    type: str
    host: str
    port: int = 9200
    index: Optional[str] = None
    auth_method: Optional[str] = None
    auth_type: Optional[Literal["basic", "apikey"]] = "basic"
    user: Optional[str] = None
    password: Optional[str] = None
    use_ssl: bool = True
    verify_certs: bool = False
    # Elasticsearch specific
    use_cloud_id: bool = False
    cloud_id: Optional[str] = None
    api_key: Optional[str] = None


import json
from datetime import datetime

from opensearchpy import OpenSearch, helpers

from shraga_common import ShragaConfig


class IngestUtils:

    @staticmethod
    def create_index(client, index):
        if client.indices.exists(index=index):
            doc_count = client.count(index=index).get("count")
            print(f"index {index} already exists, contains {doc_count} docs.")
        else:
            print(f"Creating index {index}...")
            client.indices.create(index=index)
            print(f"Index {index} created.")

    @staticmethod
    def create_index_template(client, template_name, template_body):
        if client.indices.exists_template(template_name):
            print(f"Template {template_name} already exists. Updating template...")
            client.indices.put_index_template(name=template_name, body=template_body)
            print(f"Template {template_name} updated.")
        else:
            print(f"Creating template {template_name}...")
            client.indices.put_index_template(name=template_name, body=template_body)
            print(f"Template {template_name} created.")

    @staticmethod
    def create_component_template(client, template_name, template_body):
        if client.indices.exists_template(template_name):
            print(
                f"Component template {template_name} already exists. Updating template..."
            )
            client.cluster.put_component_template(
                name=template_name, body=template_body
            )
            print(f"Component template {template_name} updated.")
        else:
            print(f"Creating component template {template_name}...")
            client.cluster.put_component_template(
                name=template_name, body=template_body
            )
            print(f"Component template {template_name} created.")

    @staticmethod
    def get_client(shraga_config: ShragaConfig):
        opens = shraga_config.get("retrievers.opensearch")
        host = opens.get("host")
        port = opens.get("port")
        auth = (opens.get("user"), opens.get("password"))
        index = opens.get("index")
        verify_certs = opens.get("verify_certs")
        print(f"Connecting to {host}:{port}/{index}...")
        return OpenSearch(
            hosts=[{"host": host, "port": port}],
            http_compress=True,  # enables gzip compression for request bodies
            http_auth=auth,
            use_ssl=True,
            verify_certs=verify_certs,
        )

    @staticmethod
    def get_client_elasticsearch(shraga_config: ShragaConfig):
        opens = shraga_config.get("retrievers.elasticsearch")
        host = opens.get("host")
        port = opens.get("port")
        auth = (opens.get("user"), opens.get("password"))
        index = opens.get("index")
        verify_certs = opens.get("verify_certs")
        use_ssl = opens.get("use_ssl")
        print(f"Connecting to {host}:{port}/{index}...")
        return OpenSearch(
            hosts=[{"host": host, "port": port}],
            http_compress=True,  # enables gzip compression for request bodies
            http_auth=auth,
            use_ssl=use_ssl,
            verify_certs=verify_certs,
        )

    @staticmethod
    def ingest(client, docs):
        succeeded = []
        failed = []

        for success, item in helpers.parallel_bulk(
            client,
            docs,
            max_chunk_bytes=1 * 1024 * 1024,
            request_timeout=30,
            raise_on_error=False,
        ):
            if success:
                succeeded.append(item)
                if len(succeeded) % 5000 == 0:
                    print(f"Bulk-inserted {len(succeeded)} items (bulk).")
            else:
                failed.append(item)

        if len(failed) > 0:
            print(f"There were {len(failed)} errors:")
            timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            with open(f"failed-{timestamp}.json", "w", encoding="utf-8-sig") as f:
                for item in failed:
                    f.write(json.dumps(item) + "\n")
        if len(succeeded) > 0:
            print(f"Bulk-inserted {len(succeeded)} items (parallel_bulk).")

    @staticmethod
    def read_json_from_gzip_file(f):
        for line in f:
            line = line.decode(encoding="UTF-8")
            yield json.loads(line)

    @staticmethod
    def count_docs(data_provider):
        count = 0
        count_by_id = {}
        for item in data_provider:
            count += 1

            item_id = item["system_id"] if "system_id" in item else "unknown"
            if item_id not in count_by_id:
                count_by_id[item_id] = 0
            count_by_id[item_id] += 1

            if count_by_id[item_id] > 1:
                print(f"Item {item_id} appears {count_by_id[item_id]} times.")

            if count % 10000 == 0:
                print(f"Processed {count} items.")

        print("Total:", count)
        for k, v in count_by_id.items():
            if v > 1:
                print(f"{k}: {v}")
        return count, count_by_id



CHAT_HISTORY_PROMPT = """

You are a helpful assistant.
You are given a question and your task is to answer questions.
You are also given the previous chat history.
Do not disclose the fact you are using a chat history. Answer as if you are a human continuing a conversation.
Base your answer on your own knowledge and the history of the conversation.

Question: 
{question}

Chat history:
{formatted_history}

Answer:
"""



from .bedrock_service import BedrockService
from .common import LLMModelResponse
from .llm_service import LlmModelProvider, LLMService, LLMServiceOptions
from .openai_service import OpenAIService

__all__ = [
    "BedrockService",
    "OpenAIService",
    "LLMService",
    "LLMServiceOptions",
    "LlmModelProvider",
    "LLMModelResponse",
]



import asyncio
from abc import abstractmethod
from typing import Any, Optional

from pydantic import BaseModel


class BaseEmbedderGenerateOptions(BaseModel):
    embed_model_name: Optional[str]
    task: Optional[str]


class BaseEmbedder:
    @abstractmethod
    async def generate_vector(
        self, text: str, extra_options: Optional[BaseEmbedderGenerateOptions] = None
    ) -> Any:
        pass

    async def generate_vectors(
        self,
        texts: list[str],
        extra_options: Optional[BaseEmbedderGenerateOptions] = None,
    ) -> Any:
        tasks = [
            self.generate_vector(
                q,
                extra_options,
            )
            for q in texts
        ]
        return await asyncio.gather(*tasks)



import json
import logging
from typing import Any, Dict, Optional, Tuple, Union

import boto3
from boto3.exceptions import Boto3Error

from shraga_common import ShragaConfig, LLMServiceUnavailableException
from .base_embedder import BaseEmbedder, BaseEmbedderGenerateOptions

# Configure logging
logger = logging.getLogger(__name__)


class BedrockEmbedder(BaseEmbedder):
    # According to cohere documentation, the maximum recomended input size is 512 tokens
    # https://docs.cohere.com/v2/reference/embed
    # The cohere embed method supports input of up to 2048 *characters*
    # See experiments/nevo/max_chunk_size.ipynb for more details
    MAX_INPUT_SIZE = 2048

    def __init__(
        self,
        shraga_config: Optional[ShragaConfig] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        profile_name: Optional[str] = None,
        region_name: str = "us-east-1",
    ):

        if shraga_config:
            aws_access_key_id = aws_access_key_id or shraga_config.get(
                "aws.access_key_id"
            )
            aws_secret_access_key = aws_secret_access_key or shraga_config.get(
                "aws.secret_access_key"
            )
            profile_name = profile_name or shraga_config.get("aws.profile")
        if aws_access_key_id == "":
            aws_access_key_id = None
        if aws_secret_access_key == "":
            aws_secret_access_key = None
        self.bedrock_client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
        )

    def get_endpoint_name_and_body(
        self, model_name: str, text: str, task="search_document"
    ) -> Tuple[str, Dict[str, Union[str, Any]]]:
        model_configs = {
            "cohere-multilingual": (
                "cohere.embed-multilingual-v3",
                {"texts": [text], "input_type": task, "truncate": "END"},
            ),
            "cohere": (
                "cohere.embed-english-v3",
                {"texts": [text], "input_type": task, "truncate": "END"},
            ),
            "titan": (
                "amazon.titan-embed-text-v2:0",
                {
                    "inputText": text,
                },
            ),
        }

        model_name_lower = model_name.lower()
        if model_name_lower in model_configs:
            return model_configs[model_name_lower]
        else:
            raise ValueError(f"Unsupported model type: {model_name}")

    async def generate_vector(
        self, text: str, extra_options: Optional[BaseEmbedderGenerateOptions] = None
    ) -> Any:
        try:
            if not extra_options:
                extra_options = {}

            if len(text) > self.MAX_INPUT_SIZE:
                logger.warning(
                    "Text is too long (%d) characters). Truncating to %d characters.",
                    len(text),
                    self.MAX_INPUT_SIZE,
                )
                text = text[: self.MAX_INPUT_SIZE]

            task = extra_options.get("task", "search_document")
            model_name = extra_options.get("embed_model_name", "cohere")
            model_id, body = self.get_endpoint_name_and_body(model_name, text, task)
            response = self.bedrock_client.invoke_model(
                modelId=model_id,
                body=json.dumps(body),
                accept="application/json",
                contentType="application/json",
            )

            response_body = response["body"].read().decode("utf-8")
            embeddings = json.loads(response_body).get("embeddings")
            if isinstance(embeddings, list):
                v = embeddings[0] if embeddings else None
                if not v:
                    raise LLMServiceUnavailableException("Empty embedding")
                return v
            else:
                raise LLMServiceUnavailableException("Unexpected embedding format")
        except Boto3Error as e:
            raise LLMServiceUnavailableException("Bedrock error", e)
        except json.JSONDecodeError as e:
            raise LLMServiceUnavailableException("Failed to decode embedding", e)
        except Exception as e:
            raise LLMServiceUnavailableException("Unexpected error", e)



import platform
import socket
import subprocess

from user_agents import parse

from shraga_common.shraga_config import ShragaConfig
from shraga_common.utils import is_prod_env


def get_git_commit():
    try:
        # Run git command to get the short commit hash
        commit_hash = (
            subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
            .strip()
            .decode("utf-8")
        )
        return commit_hash
    except subprocess.CalledProcessError:
        return None


def get_git_branch():
    try:
        # Run git command to get the current branch name
        branch_name = (
            subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
            .strip()
            .decode("utf-8")
        )
        return branch_name
    except subprocess.CalledProcessError:
        return None


def get_user_agent_info(user_agent: str):
    if not user_agent:
        return None

    ua = parse(user_agent)
    device_type = (
        "pc"
        if ua.is_pc
        else "tablet" if ua.is_tablet else "mobile" if ua.is_mobile else "other"
    )

    return {
        "browser": ua.browser.family,
        "browser_version": ua.browser.version_string,
        "os": ua.os.family,
        "os_version": ua.os.version_string,
        "device": ua.device.family,
        "device_type": device_type,
        "device_brand": ua.device.brand,
        "device_model": ua.device.model,
        "original": user_agent,
    }


def get_platform_info():
    return {
        "system": platform.system(),
        "node_name": platform.node(),
        "release": platform.release(),
        "version": platform.version(),
        "architecture": platform.architecture(),
        "processor": platform.processor(),
        "platform_details": platform.platform(),
        "python_version": platform.python_version(),
        "machine_name": socket.gethostname(),
        "git_commit": get_git_commit(),
        "git_branch": get_git_branch(),
    }


def get_config_info(shraga_config: ShragaConfig):
    return {
        "debug.enabled": shraga_config.get("debug.enabled"),
        "ui.default_flow": shraga_config.get("ui.default_flow"),
        "prod": is_prod_env(),
    }



import asyncio
import gzip
import json
import multiprocessing
import os
import platform
import queue
import uuid
from collections.abc import Iterable
from multiprocessing import Process, Queue
from typing import Optional

from pydantic import BaseModel

from shraga_common.preprocessing.doc_handler import DocHandler


class PreprocessingRunnerPreferences(BaseModel):
    parallelism: bool = True
    number_of_processes: int = 20
    number_of_push_threads: int = 1
    queue_timeout: int = 60
    max_queue_size: int = 10_000
    output_file_prefix: Optional[str] = ""


class PreprocessingRunner:
    doc_handler: DocHandler = None
    out_path: str = None
    data_provider: Iterable = ([],)
    config: PreprocessingRunnerPreferences = None

    def __init__(
        self,
        doc_handler: DocHandler,
        data_provider: Iterable,
        out_path: str = None,
        config: Optional[PreprocessingRunnerPreferences] = None,
    ):
        self.data_provider = data_provider
        self.doc_handler = doc_handler
        self.out_path = out_path
        self.config = config or PreprocessingRunnerPreferences()
        is_first_run = not multiprocessing.get_start_method(allow_none=True)
        if platform.system() == "Darwin" and is_first_run:
            multiprocessing.set_start_method("fork")

    def run_async_job(self, tasks_to_accomplish, tasks_that_are_done):
        try:
            asyncio.run(self.do_job(tasks_to_accomplish, tasks_that_are_done))
        except KeyboardInterrupt:
            print("Job interrupted by user. Exiting gracefully.")

    def get_filename(self):
        output_file_prefix = ""
        if self.config.output_file_prefix:
            output_file_prefix = f"{self.config.output_file_prefix}_"
        return f"{self.out_path}/{output_file_prefix}{uuid.uuid4().hex}.jsonl.gz"

    async def do_job(self, tasks_to_accomplish, tasks_that_are_done):
        print(f"[{os.getpid()}] Started")
        while True:
            try:
                task = tasks_to_accomplish.get(block=True, timeout=5)
            except queue.Empty:
                print(f"[{os.getpid()}] Queue is empty")
            else:
                if task is None:
                    print(f"[{os.getpid()}] Got sentinel")
                    tasks_that_are_done.put(None)
                    # use sentinel pattern
                    break
                try:
                    async for result in self.doc_handler.handle_document(task):
                        if not result:
                            break
                        tasks_that_are_done.put(result)
                except Exception as e:
                    print(f"[{os.getpid()}] Error: {e}")
                    continue

        print(f"[{os.getpid()}] Stopped")
        return True

    def push_data(self, tasks_that_are_done):
        print("[Push Thread] Started")
        processed_ids = set()
        total_chunks = 0
        sentinel_count = 0

        if not os.path.exists(self.out_path):
            print(f"[Push Thread] Creating output directory: {self.out_path}")
            os.makedirs(self.out_path)

        filename = self.get_filename()
        with gzip.open(filename, "wt", encoding="utf-8") as gzfile:
            while True:
                try:
                    chunk = tasks_that_are_done.get(block=True, timeout=5)
                except queue.Empty:
                    print("[Push Thread] Queue is empty")
                else:
                    if chunk is None:
                        print("[Push Thread] Got sentinel")
                        sentinel_count += 1
                        if sentinel_count == self.config.number_of_processes:
                            print("[Push Thread] All processes finished")
                            break
                    else:
                        gzfile.write(json.dumps(chunk, ensure_ascii=False, indent=None))
                        gzfile.write("\n")

                        processed_ids.add(chunk.get("id"))
                        total_chunks += 1

                        if total_chunks and total_chunks % 1000 == 0:
                            print(
                                f"[Push Thread] Pushed (total {total_chunks} chunks, {len(processed_ids)} items)"
                            )

        print("[Push Thread] Stopped")
        return True

    async def execute(self):
        tasks_to_accomplish = Queue()
        tasks_that_are_done = Queue()

        processes = []
        push_threads = []

        if not self.config.parallelism:
            count = 0
            for data in self.data_provider:
                if isinstance(data, list):
                    for d in data:
                        tasks_to_accomplish.put(d)
                        count += 1
                else:
                    tasks_to_accomplish.put(data)
                    count += 1

                if count >= self.config.max_queue_size:
                    await self.do_job(tasks_to_accomplish, tasks_that_are_done)
                    self.push_data(tasks_that_are_done)
                    count = 0

            await self.do_job(tasks_to_accomplish, tasks_that_are_done)
            self.push_data(tasks_that_are_done)
        else:
            for _ in range(self.config.number_of_processes):
                p = Process(
                    target=self.run_async_job,
                    args=(tasks_to_accomplish, tasks_that_are_done),
                )
                processes.append(p)
                p.start()

            for _ in range(self.config.number_of_push_threads):
                push_thread = Process(
                    target=self.push_data, args=(tasks_that_are_done,)
                )
                push_threads.append(push_thread)
                push_thread.start()

            count = 0
            for data in self.data_provider:
                count += 1
                if count >= self.config.max_queue_size:
                    while not tasks_to_accomplish.empty():
                        print(
                            "Queue is full. Waiting for tasks to be processed before loading more..."
                        )
                        await asyncio.sleep(1)
                    count = 0

                if isinstance(data, list):
                    for d in data:
                        tasks_to_accomplish.put(d)
                else:
                    tasks_to_accomplish.put(data)

            # send shut down signal to all task workers
            for _ in range(self.config.number_of_processes):
                tasks_to_accomplish.put(None)

            for p in processes:
                p.join()
            for p in push_threads:
                p.join()



import binascii

import requests
from starlette.authentication import (AuthCredentials, AuthenticationBackend,
                                     AuthenticationError)

from .user import ShragaUser


class MicrosoftAuthBackend(AuthenticationBackend):
    async def authenticate(self, conn):
        if "user" in conn and conn.user.is_authenticated:
            return AuthCredentials(["authenticated"]), conn.user

        if "Authorization" not in conn.headers:
            raise AuthenticationError("Unauthenticated")

        auth = conn.headers["Authorization"]
        try:
            scheme, token = auth.split()
            if scheme.lower() != "microsoft":
                return
            response = requests.get(
                "https://graph.microsoft.com/v1.0/me",
                headers={"Authorization": f"Bearer {token}"},
                timeout=10
            )
            if response.status_code != 200:
                raise AuthenticationError("Invalid Microsoft OAuth token")
            user_info = response.json()
        except (ValueError, UnicodeDecodeError, binascii.Error) as exc:
            raise AuthenticationError("Invalid Microsoft OAuth token") from exc

        username = user_info.get("mail") or user_info.get("userPrincipalName")
        
        metadata = {
            "display_name": user_info.get("displayName"),
            "email": user_info.get("mail") or user_info.get("userPrincipalName"),
            "user_id": user_info.get("id"),
            "auth_type": "microsoft",
        }
        
        # Remove None values
        metadata = {k: v for k, v in metadata.items() if v is not None}
            
        user = ShragaUser(
            username=username,
            metadata=metadata
        )

        return AuthCredentials(["authenticated"]), user



from typing import Optional
from fastapi import APIRouter, HTTPException, Query
from fastapi.responses import JSONResponse

from ..services import report_service
from ..models.report_request import ReportRequest

router = APIRouter()

@router.post("/export")
async def generate_report(request: ReportRequest) -> JSONResponse:
    try:
        data = await report_service.generate_report(
            request.report_type,
            request.start,
            request.end,
            request.filters
        )
        
        return JSONResponse(
            status_code=200,
            content={
                "report_type": request.report_type,
                "total_records": len(data),
                "data": data
            }
        )
        
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal server error: {str(e)}")



from .flow_base import FlowBase
from .flow_response import FlowResponse
from .flow_run_request import FlowRunRequest, HistoryMessage
from .flow_stats import FlowStats
from .retrieval_result import RetrievalResult

__all__ = [
    "FlowBase",
    "FlowStats",
    "FlowResponse",
    "FlowRunRequest",
    "RetrievalResult",
    "HistoryMessage",
]



import unittest
from unittest.mock import patch, MagicMock, AsyncMock
from shraga_common.services.bedrock_service import BedrockService, InvokeConfig, LLMServiceUnavailableException
import json
from json import JSONDecodeError

class TestBedrockService(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.service = BedrockService()
        self.system_prompt = ["system"]
        self.prompt = "prompt"
        self.tool_config = {}

    async def test_invoke_converse_model_invalid_json_triggers_retry(self):
        # Simulate a response with invalid JSON in the text field
        mock_boto = MagicMock()
        
        mock_boto.converse.side_effect = [
            # First call - invalid JSON
            {
                "metrics": {"latencyMs": 1},
                "usage": {"inputTokens": 1, "outputTokens": 1, "totalTokens": 2},
                "output": {"message": {"content": [{"text": '{"aaaaaaa": )'}]}},
            },
            # Second call (retry) - valid JSON
            {
                "metrics": {"latencyMs": 1},
                "usage": {"inputTokens": 1, "outputTokens": 1, "totalTokens": 2},
                "output": {"message": {"content": [{"text": '{"ok": true}'}]}},
            }
        ]
        self.service.boto = mock_boto
        options = InvokeConfig(model_id="sonnet_3", parse_json=True, is_retry=False)

        # Patch self.service.invoke_converse_model to call the real method, but monitor calls
        with patch.object(BedrockService, "invoke_converse_model", wraps=self.service.invoke_converse_model) as spy_invoke:
            result = await self.service.invoke_converse_model(self.system_prompt, self.prompt, self.tool_config, options)
            # The spy should have been called twice: once for the initial call, once for the retry
            self.assertEqual(spy_invoke.call_count, 2)
            # After retry, is_retry should be True
            self.assertTrue(options.is_retry)
            # The result should be the successful retry result
            self.assertEqual(result.json, {"ok": True})
            
            
    async def test_invoke_converse_model_invalid_json_fails_after_retry(self):
        # Simulate a response with invalid JSON in the text field
        mock_boto = MagicMock()

        mock_boto.converse.side_effect = [
            # First call - invalid JSON
            {
                "metrics": {"latencyMs": 1},
                "usage": {"inputTokens": 1, "outputTokens": 1, "totalTokens": 2},
                "output": {"message": {"content": [{"text": '{"aaaaaaa": )'}]}},
            },
            # Second call (retry) - valid JSON
            {
                "metrics": {"latencyMs": 1},
                "usage": {"inputTokens": 1, "outputTokens": 1, "totalTokens": 2},
                "output": {"message": {"content": [{"text": '{"aaaaaaa": )'}]}},
            }
        ]
        self.service.boto = mock_boto
        options = InvokeConfig(model_id="sonnet_3", parse_json=True, is_retry=False)

        # Patch self.service.invoke_converse_model to call the real method, but monitor calls
        with patch.object(BedrockService, "invoke_converse_model", wraps=self.service.invoke_converse_model) as spy_invoke:
            # Assert that JSONDecodeError is raised 
            with self.assertRaises(JSONDecodeError):
                await self.service.invoke_converse_model(self.system_prompt, self.prompt, self.tool_config, options)

    async def test_invoke_converse_model_valid_json_no_retry(self):
        # Simulate a response with valid JSON in the text field
        mock_boto = MagicMock()
        mock_boto.converse.return_value = {
            "metrics": {"latencyMs": 1},
            "usage": {"inputTokens": 1, "outputTokens": 1, "totalTokens": 2},
            "output": {"message": {"content": [{"text": '{"ok": true}'}]}},
        }
        self.service.boto = mock_boto
        options = InvokeConfig(model_id="sonnet_3", parse_json=True, is_retry=False)
        
        # No need to patch json.loads - the real one should work fine with valid JSON
        with patch.object(BedrockService, "invoke_converse_model", wraps=self.service.invoke_converse_model) as spy_invoke:
            result = await self.service.invoke_converse_model(self.system_prompt, self.prompt, self.tool_config, options)
            
            # Assertions
            self.assertEqual(spy_invoke.call_count, 1)
            self.assertFalse(options.is_retry)
            self.assertEqual(result.json, {"ok": True})

    async def test_invoke_converse_model_parse_json_false_does_not_call_json_loads(self):
        # Simulate a response with any text
        mock_boto = MagicMock()
        mock_boto.converse.return_value = {
            "metrics": {"latencyMs": 1},
            "usage": {"inputTokens": 1, "outputTokens": 1, "totalTokens": 2},
            "output": {"message": {"content": [{"text": '{"ok": true}'}]}},
        }
        self.service.boto = mock_boto
        options = InvokeConfig(model_id="sonnet_3", parse_json=False, is_retry=False)

        with patch("json.loads") as mock_json_loads:
            result = await self.service.invoke_converse_model(self.system_prompt, self.prompt, self.tool_config, options)
            mock_json_loads.assert_not_called()
            self.assertIsNone(result.json)

    async def test_invoke_converse_model_llm_service_unavailable_exception_no_retry(self):
        # Simulate a response that raises LLMServiceUnavailableException
        mock_boto = MagicMock()
        mock_boto.converse.side_effect = LLMServiceUnavailableException("Bedrock error")
        self.service.boto = mock_boto
        options = InvokeConfig(model_id="sonnet_3", parse_json=True, is_retry=False)

        with patch("json.loads") as mock_json_loads:
            with self.assertRaises(LLMServiceUnavailableException):
                await self.service.invoke_converse_model(self.system_prompt, self.prompt, self.tool_config, options)
            # Ensure json.loads was never called
            mock_json_loads.assert_not_called()
            # Ensure retry logic was not triggered
            self.assertFalse(options.is_retry)

if __name__ == "__main__":
    unittest.main()



import argparse
import asyncio
import base64
import hashlib
import json
import os
import traceback
from abc import abstractmethod
from datetime import datetime, timezone
from typing import Dict, List, Optional

import yaml
from pydantic import ValidationError

from shraga_common import ShragaConfig
from shraga_common.flows.builtin.evaluation_model import (EvaluationModel,
                                                          EvaluationScenario)
from shraga_common.logger import get_config_info, get_platform_info
from shraga_common.models import FlowBase, FlowResponse, FlowRunRequest
from shraga_common.prompts import PART_EVALUATION_PROMPT
from shraga_common.retrievers import get_client
from shraga_common.services import BedrockService, LLMService


class EvaluationFlow(FlowBase):
    llmservice: LLMService = None
    listed = False
    files = []
    output_file_raw_results = None

    def __init__(self, config: ShragaConfig, flows: Optional[dict] = None):
        super().__init__(config, flows)
        self.es_client = get_client(config)
        self.index_name = config.get("evaluation.index")

        # set the evaluation id to the current date YY-MM-dd + a UUID
        platform_info = get_platform_info()
        ts = datetime.now()
        ts_str = ts.strftime("%Y-%m-%d-%H-%M-%S-%f")[:-3]
        machine = platform_info.get("machine_name", "unknown")[:3]
        self.evaluation_ts = ts
        self.evaluation_id = f"{ts_str}-{machine}"
        print(f"Evaluation ID: {self.evaluation_id}")
        self.runtime = datetime.now(timezone.utc).isoformat()

    @staticmethod
    def parse_args():
        parser = argparse.ArgumentParser(description="Run the Nevo Evaluation Flow")
        parser.add_argument(
            "--test_count", type=int, default=3, help="Number of scenarios to run"
        )
        parser.add_argument(
            "--input_file", type=str, default=None, help="Input file name to run"
        )
        parser.add_argument(
            "--input_files",
            type=str,
            default=None,
            help="List of input file name to run",
        )
        parser.add_argument(
            "--tag", type=str, default=None, help="Filter data by specific tag"
        )

        parser.add_argument(
            "--run_only",
            action=argparse.BooleanOptionalAction,
            default=False,
            help="Don't evaluate results",
        )
        args = parser.parse_args()
        return vars(args)

    def use_args(self, request: FlowRunRequest, base_path: str):
        if request.preferences.get("input_file"):
            file_names = [request.preferences.get("input_file")]
        elif request.preferences.get("input_files"):
            file_names = request.preferences.get("input_files")
        else:
            file_names = self.files

        request.preferences["input_files"] = [
            os.path.join(base_path, "evaluation", filename) for filename in file_names
        ]

    def get_md5_hash(self, text: str):
        hasher = hashlib.sha1(text.encode(encoding="UTF-8", errors="strict"))
        return base64.urlsafe_b64encode(hasher.digest()).decode("utf-8-sig")

    def get_routing_info(self, scenario_answer: dict):
        return scenario_answer.get("routing.payload")

    @abstractmethod
    def is_key_doc_id_in_retrieval_results(
        self, scenario_answer: dict, scenario: EvaluationScenario
    ):
        pass

    def format_docs(self, item):
        return {
            "id": item.id,
            "document_id": item.document_id,
            "date": item.date,
            "title": item.title,
            "incident_type": item.extra["incident_type"],
            "link": item.link,
            "score": item.score,
        }

    async def run_scenario(
        self, scenario: EvaluationScenario, preferences: EvaluationModel
    ):
        self.trace(f"Running scenario {scenario.question}")
        start_time = datetime.now()
        evaluated_flow_preferences = preferences.evaluated_flow_preferences or {}
        self.trace(f"Overriding flow preferences: {evaluated_flow_preferences}")
        flow_run_request = FlowRunRequest(
            flows=self.flows,
            question=scenario.question,
            preferences={
                **evaluated_flow_preferences,
                "last_lookup_date": preferences.last_lookup_date,
            },
        )
        scenario_answer = await self.execute_another_flow_by_id(
            preferences.flow_id, flow_run_request
        )

        if not scenario_answer or not scenario_answer.payload:
            self.trace(f"Failed to run scenario {scenario.question}")
            return None

        run_time = datetime.now() - start_time
        answer = scenario_answer.payload.get("answer", "")

        testcase = scenario.dict()
        # hash the question to track changes over time
        testcase["id"] = self.get_md5_hash(scenario.question)

        # check if the document id or report url is in the retrieval results
        contains_key_doc_id = self.is_key_doc_id_in_retrieval_results(
            scenario_answer, scenario
        )

        # Calculate Average Precision
        avg_precision = self.average_precision(scenario_answer, scenario)

        missing_doc_id = contains_key_doc_id is None
        no_answer = answer.startswith("I'm sorry")

        self.trace(f"---\nAnswer: {answer}\n---")
        return {
            "testcase": testcase,
            "generated_answer": answer,
            "evaluation": {
                "missing_doc_id": missing_doc_id,
                "contains_key_doc_id": contains_key_doc_id,
                "average_precision": avg_precision,
                "no_answer": no_answer,
                "run_time": run_time.seconds,
            },
            "metadata": {
                "routing": self.get_routing_info(scenario_answer),
                "retrieved_documents": [
                    self.format_docs(item) for item in scenario_answer.retrieval_results
                ],
                "trace": scenario_answer.trace,
                "exec_stats": [s.dict() for s in scenario_answer.stats],
            },
        }
    
    def average_precision(self, scenario_answer, scenario: EvaluationScenario):
        hits = 0
        sum_precisions = 0.0
        retrieved_docs = [doc.title for doc in scenario_answer.retrieval_results]
        relevant_docs = scenario.metadata.get('document_ids')

        for i, doc_id in enumerate(retrieved_docs):
            if doc_id in relevant_docs:
                hits += 1
                precision_at_i = hits / (i + 1)
                sum_precisions += precision_at_i

        if hits == 0 or len(relevant_docs) == 0:
            return 0.0
        return sum_precisions / len(relevant_docs)

    def get_eval_prompt(
        self, testcase: dict = None, response_answer: str = None, _=None
    ):
        expected_answer = testcase.get("answer", None)
        return PART_EVALUATION_PROMPT.format(
            question=testcase.get("question", ""),
            expected_answer=expected_answer,
            answer=response_answer,
        )

    async def evaluate_results(self, results: List, preferences: EvaluationModel):
        self.llmservice = self.llmservice or BedrockService(self.config)

        if not results:
            print("No results to evaluate")
            return

        self.trace(f"Evaluating {len(results)} scenarios")
        correct_count = 0
        no_answer_count = 0
        partial_correct_count = 0
        average_run_time = 0
        found_key_doc_id = 0
        no_key_doc_id = 0
        err_count = 0
        ap_scores = []

        # Lock to safely update counters in parallel
        lock = asyncio.Lock()

        async def evaluate_single_result(result):
            nonlocal correct_count, no_answer_count, partial_correct_count, average_run_time, found_key_doc_id, no_key_doc_id

            if not result:
                self.trace("No result to evaluate")
                return

            testcase = result.get("testcase", {})
            response_answer = result.get("generated_answer", None)
            average_run_time += result.get("evaluation", {}).get("run_time", 0)
            expected_answer = testcase.get("answer", None)

            ap_scores.append(result["evaluation"].get("average_precision"))

            if result["evaluation"].get("contains_key_doc_id"):
                found_key_doc_id += 1
            elif result["evaluation"].get("contains_key_doc_id") is None:
                no_key_doc_id += 1

            if expected_answer:
                prompt = self.get_eval_prompt(testcase, response_answer, preferences)
                eval_resp = await self.llmservice.invoke_model(
                    prompt, {"model_id": "sonnet_3_5_v2"}
                )

                if not eval_resp:
                    self.trace("No response from the model")
                    return

                try:

                    result["evaluation"] = {
                        **result["evaluation"],
                        **json.loads(eval_resp.text, strict=False),
                    }
                except Exception as _:
                    result["evaluation"] = {
                        **result["evaluation"],
                        "error": "Error parsing evaluation response",
                    }
                    print("Error parsing evaluation response", eval_resp.text)

                # Update counters based on evaluation results
                async with lock:
                    correctness = result["evaluation"].get("correctness")
                    if correctness == "correct":
                        correct_count += 1
                    elif correctness == "partially correct":
                        partial_correct_count += 1
                    if result["evaluation"].get("no_answer"):
                        no_answer_count += 1
                        result["evaluation"]["correctness"] = "no answer"

                    if result["evaluation"].get("error"):
                        err_count += 1

                    self.log_eval_result(result, preferences)
            else:
                self.log_eval_result(result, preferences)

        # Run evaluations in parallel for each result
        tasks = [evaluate_single_result(result) for result in results]
        await asyncio.gather(*tasks)

        # Calculate averages after all tasks complete
        if average_run_time > 0:
            average_run_time /= len(results)

        # Compile evaluation data
        data = {
            "run_date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "preferences": preferences.dict(),
            "stats": {
                "total_scenarios": len(results),
                "error_count": err_count,
                "total_found_key_doc_id": found_key_doc_id,
                "no_key_doc_id": no_key_doc_id,
                "not_found_key_doc_id": len(results) - found_key_doc_id - no_key_doc_id,
                "correct_count": correct_count,
                "no_answer_count": no_answer_count,
                "partial_correct_count": partial_correct_count,
                "correct_percentage": round(correct_count / len(results) * 100),
                "partial_correct_percentage": round(
                    partial_correct_count / len(results) * 100
                ),
                "average_run_time": average_run_time,
                "mean_average_precision": round(sum(ap_scores) / len(ap_scores), 3) if ap_scores else 0
            },
            "results": results,
        }
        return data

    def log_eval_result(self, result: dict, preferences: EvaluationModel):
        if not self.es_client:
            return
        
        testcase = result.get("testcase", {})
        response_answer = result.get("generated_answer", None)

        o = {}
        o["timestamp"] = self.runtime
        o["question"] = testcase.get("question")
        o["question_hash"] = testcase.get("id")
        o["gen_answer"] = response_answer
        o["evaluation"] = result.get("evaluation", {})
        o["config"] = get_config_info(self.config)
        o["evaluation_ts"] = self.evaluation_ts
        o["evaluation_id"] = self.evaluation_id
        o["platform"] = get_platform_info()
        o["preferences"] = {
            **preferences.dict(),
            "input_files": [f.split("/")[-1] for f in preferences.input_files],
        }
        docs = result.get("metadata", {}).get("retrieved_documents", [])
        for doc in docs:
            del doc["date"]
        o["retrieved_documents"] = docs
        if o["evaluation"].get("correctness") != "correct":
            o["trace"] = result.get("metadata", {}).get("trace", [])
            o["routing"] = result.get("metadata", {}).get("routing", [])

        if not self.index_name:
            raise ValueError(f"Index '{self.index_name}' is not correctly defined in the config")
        
        self.es_client.index(index=self.index_name, body=o)

    def reached_test_count(self, results: List, preferences: EvaluationModel):
        test_count = preferences.test_count
        if test_count <= 0:
            return False
        return len(results) >= test_count

    def get_scenarios(self, dataset: Dict):
        return dataset.get("scenarios", [])

    async def run_scenarios(self, dataset: Dict, preferences: EvaluationModel):
        scenarios = self.get_scenarios(dataset)

        max_tests = min(len(scenarios), preferences.test_count)
        self.trace(f"Running {max_tests} scenarios")

        results = []
        max_concurrency = preferences.max_concurrent_tests
        semaphore = asyncio.Semaphore(max_concurrency)
        lock = asyncio.Lock()
        test_counter = 0  # Counter for completed tests

        async def run_single_scenario(scenario):
            nonlocal test_counter
            async with semaphore:
                # Check if we've reached the maximum count
                async with lock:
                    if max_tests > 0 and test_counter >= max_tests:
                        return None
                    else:
                        test_counter += 1

                try:
                    scenario_object = EvaluationScenario(**scenario)
                    result = await self.run_scenario(scenario_object, preferences)
                    results.append(result)
                    self.write_test_output(result, self.output_file_raw_results)

                    return result
                except ValidationError as e:
                    self.trace(e)
                    self.trace(f"Invalid scenario: {scenario.get('question')}")
                except Exception as e:
                    print(traceback.print_exc())
                    self.trace(e)
                    self.trace(f"Error running scenario: {scenario.get('question')}")
                return None

        # Start all scenarios as tasks
        tasks = [run_single_scenario(scenario) for scenario in scenarios]
        await asyncio.gather(*tasks, return_exceptions=True)
        for result in results:
            if isinstance(result, Exception):
                print(f"Task failed with exception: {result}")

        return results[:max_tests]

    def write_eval_output(self, data: dict, output_file: str):
        with open(output_file, "w", encoding="utf-8-sig") as stream:
            yaml.safe_dump(data, stream, sort_keys=False, allow_unicode=True)

    def write_test_output(self, test_data: dict, output_file: str):
        with open(output_file, "a", encoding="utf-8-sig") as stream:
            if test_data:
                yaml.safe_dump([test_data], stream, sort_keys=False, allow_unicode=True)

    async def execute(self, request: FlowRunRequest) -> FlowResponse:
        self.llmservice = self.llmservice or BedrockService(self.config)
        preferences = EvaluationModel(**request.preferences)

        timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
        fname = f"{preferences.flow_id}-{timestamp}-eval.yaml"
        output_file = os.path.join(os.getcwd(), "output", fname)
        
        # ensure folder exists
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        
        self.output_file_raw_results = os.path.join(
            os.getcwd(), "output", "raw-" + fname
        )

        self.write_test_output(None, self.output_file_raw_results)
        self.trace(f"Storing raw results to {self.output_file_raw_results}")
        self.trace(f"Storing evaluation results to {output_file}")

        results = []

        for input_file in preferences.input_files:
            self.trace(f"Loading the dataset {input_file}")
            with open(input_file, "r", encoding="utf-8-sig") as stream:
                try:
                    dataset = yaml.safe_load(stream)
                    preferences.last_lookup_date = dataset["api_context"].get("date")
                except yaml.YAMLError as exc:
                    self.trace(exc)

                if preferences.tag:
                    dataset["scenarios"] = [
                        item
                        for item in dataset["scenarios"]
                        if "tags" in item["metadata"]
                        and preferences.tag in item["metadata"]["tags"]
                    ]

                dataset_results = await self.run_scenarios(dataset, preferences)
                results.extend(dataset_results)
                if self.reached_test_count(dataset_results, preferences):
                    self.trace("Max scenarios reached")
                    break

        if not preferences.run_only:
            evaluation_results = await self.evaluate_results(results, preferences)
            self.write_eval_output(evaluation_results, output_file)

        return FlowResponse(
            response_text="",
            payload={
                "output_file": output_file,
                "total_scenarios": len(results),
            },
        )

    @staticmethod
    def id():
        return "llm-evaluation-v1"

    @staticmethod
    def description():
        return "evaluate RAG results using LLMs"



import unittest
from unittest.mock import AsyncMock, patch

from flows.demo.flow_chat_demo import ChatDemoFlow

from shraga_common.models import FlowRunRequest, FlowStats, HistoryMessage
from shraga_common.services import LLMModelResponse


class TestChatDemoFlow(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):

        mock_config = {
            "region": "us-east-1",
            "model_id": "sonnet_3_5",
            "max_retries": 1,
        }

        self.flow = ChatDemoFlow(config=mock_config)

    @patch("shraga_common.services.BedrockService.invoke_model", new_callable=AsyncMock)
    async def test_execute_with_history(self, mock_invoke_model):

        mock_invoke_model.return_value = LLMModelResponse(
            text="Your name is Daniel", stats=FlowStats(flow_id=None)
        )

        history = [
            HistoryMessage(text="Hi, how are you? My name is Daniel", msg_type="user"),
            HistoryMessage(text="I'm ok, how are you?", msg_type="system"),
        ]

        request = FlowRunRequest(question="What is my name?", chat_history=history)

        result = await self.flow.execute(request)
        mock_invoke_model.assert_called_once()
        self.assertIn(
            "Daniel",
            result.payload["body"].text,
            "The answer does not contain 'Daniel'.",
        )



import os
from typing import List

from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.gzip import GZipMiddleware

from shraga_common.logger import get_git_commit
from shraga_common.models import FlowBase

from ..config import get_config, load_config
from ..middlewares import logging_middleware
from ..services import list_flows_service
from .api_app import api_app, load_api_app
from .oauth_app import load_oauth_app, oauth_app


def setup_app(config_path: str, flows: List[FlowBase]) -> FastAPI:
    shraga_config = load_config(config_path)
    
    if not shraga_config:
        exit(1)
    
    list_flows_service.register_flows(flows, shraga_config)
    load_api_app()
    load_oauth_app()

    app = FastAPI(
        title="Shraga",
        description="Shraga AI",
        openapi_url="/openapi.json",
        # version=version,
        docs_url=None,
        redoc_url=None,
    )
    app.add_middleware(GZipMiddleware, minimum_size=5000)
    app.add_middleware(BaseHTTPMiddleware, dispatch=logging_middleware)
    app.mount("/api", api_app)
    app.mount("/oauth", oauth_app)


    @app.middleware("http")
    async def version_header_middleware(request, call_next):
        response = await call_next(request)
        response.headers["x-shraga-version"] = get_git_commit() or "unknown"
        return response


    @app.get("/healthz")
    async def healthz() -> dict:
        return {"ok": True, "build": os.getenv("SHRAGA_BUILD_TAG")}


    @app.get("/auth/login_methods")
    async def login_methods() -> List[str]:
        realms = list(shraga_config.auth_realms().keys())
        if "basic" in realms and ("google" in realms or "microsoft" in realms):
            realms.remove("basic")
            return realms
        return realms
    
    if get_config("ui.enabled") != "false":
        dist_path = "frontend/dist"
        if os.path.exists(dist_path):
            # Serve the frontend
            templates = Jinja2Templates(directory=dist_path)
            app.mount(
                "/assets", StaticFiles(directory=f"{dist_path}/assets"), name="static"
            )

            @app.get("/{path:path}", response_class=HTMLResponse)
            async def serve_react_app(request: Request, path: str):
                if path.startswith("api"):
                    raise HTTPException(404)
                return templates.TemplateResponse("index.html", {"request": request})


    return app


import datetime
import unittest
from unittest.mock import MagicMock, patch

import jwt
from starlette.authentication import AuthCredentials, AuthenticationError

from shraga_common.app.auth.jwt_auth import JWTAuthBackend
from shraga_common.app.auth.user import ShragaUser
from shraga_common.shraga_config import ShragaConfig


class TestJWTAuthBackendExtended(unittest.IsolatedAsyncioTestCase):
    """Extended test suite for JWTAuthBackend class focusing on error handling and metadata."""
    
    def setUp(self):
        """Set up test fixtures before each test method."""
        self.auth_backend = JWTAuthBackend()
        
        # Test credentials
        self.test_username = "test@example.com"
        self.test_secret = "test_secret_key"
        
        # Create test JWT payload
        self.test_payload = {
            "username": self.test_username,
            "email": self.test_username,
            "user_id": "user123",
            "company": "test_company",
            "roles": ["admin", "user"]
        }
        
        # Create encoded JWT token
        self.test_token = jwt.encode(
            self.test_payload, 
            self.test_secret, 
            algorithm="HS256"
        )
        
        # Create auth header
        self.auth_header = f"Bearer {self.test_token}"
        
        # Mock connection object
        self.conn = MagicMock()
        self.conn.headers = {"Authorization": self.auth_header}
        self.conn.user = MagicMock()
        self.conn.user.is_authenticated = False

    @patch("shraga_common.app.auth.jwt_auth.get_config")
    async def test_authenticate_with_expired_token(self, mock_get_config):
        """Test authentication with expired JWT token."""
        # Create payload with expired token
        expired_payload = {
            "username": self.test_username,
            "exp": datetime.datetime.now().timestamp() - 3600  # 1 hour in the past
        }
        
        # Create token and auth header
        expired_token = jwt.encode(
            expired_payload, 
            self.test_secret, 
            algorithm="HS256"
        )
        self.conn.headers = {"Authorization": f"Bearer {expired_token}"}
        
        # Set up mock config
        mock_config = MagicMock(spec=ShragaConfig)
        mock_config.auth_realms.return_value = {
            "jwt": {"secret": self.test_secret}
        }
        mock_get_config.return_value = mock_config
        
        # Call authenticate and verify it raises error
        with self.assertRaises(jwt.ExpiredSignatureError) as context:
            await self.auth_backend.authenticate(self.conn)
        
        self.assertEqual(str(context.exception), "Signature has expired")

    @patch("shraga_common.app.auth.jwt_auth.get_config")
    async def test_authenticate_with_roles(self, mock_get_config):
        """Test authentication with roles in JWT token."""
        # Create payload with roles
        payload_with_roles = {
            "username": self.test_username,
            "email": self.test_username,
            "roles": ["admin", "user", "moderator"]
        }
        
        # Create token and auth header
        token = jwt.encode(
            payload_with_roles, 
            self.test_secret, 
            algorithm="HS256"
        )
        self.conn.headers = {"Authorization": f"Bearer {token}"}
        
        # Set up mock config
        mock_config = MagicMock(spec=ShragaConfig)
        mock_config.auth_realms.return_value = {
            "jwt": {"secret": self.test_secret}
        }
        mock_get_config.return_value = mock_config
        
        # Call authenticate
        credentials, user = await self.auth_backend.authenticate(self.conn)
        
        # Verify results
        self.assertEqual(credentials.scopes, AuthCredentials(["authenticated"]).scopes)
        self.assertEqual(user.username, self.test_username)
        self.assertIsInstance(user, ShragaUser)
        
        # Verify roles are correctly in metadata
        self.assertEqual(user.get_metadata("roles"), ["admin", "user", "moderator"])
        
        # Verify other metadata
        self.assertEqual(user.get_metadata("auth_type"), "jwt")
        self.assertEqual(user.get_metadata("email"), self.test_username)

    @patch("shraga_common.app.auth.jwt_auth.get_config")
    async def test_authenticate_with_null_values(self, mock_get_config):
        """Test authentication with null values in JWT token."""
        # Create payload with null values
        payload_with_nulls = {
            "username": self.test_username,
            "email": None,
            "user_id": "user123",
            "company": None
        }
        
        # Create token and auth header
        token = jwt.encode(
            payload_with_nulls, 
            self.test_secret, 
            algorithm="HS256"
        )
        self.conn.headers = {"Authorization": f"Bearer {token}"}
        
        # Set up mock config
        mock_config = MagicMock(spec=ShragaConfig)
        mock_config.auth_realms.return_value = {
            "jwt": {"secret": self.test_secret}
        }
        mock_get_config.return_value = mock_config
        
        # Call authenticate
        credentials, user = await self.auth_backend.authenticate(self.conn)
        
        # Verify results
        self.assertEqual(user.username, self.test_username)
        
        # Verify null values are filtered out from metadata
        self.assertEqual(user.get_metadata("auth_type"), "jwt")
        self.assertEqual(user.get_metadata("user_id"), "user123")
        self.assertIsNone(user.get_metadata("email"))
        self.assertIsNone(user.get_metadata("company"))

    @patch("shraga_common.app.auth.jwt_auth.get_config")
    async def test_authenticate_missing_jwt_secret(self, mock_get_config):
        """Test authentication with missing JWT secret in config."""
        # Set up mock config with missing secret
        mock_config = MagicMock(spec=ShragaConfig)
        mock_config.auth_realms.return_value = {
            "jwt": {}  # Missing secret
        }
        mock_get_config.return_value = mock_config
        
        with self.assertRaises(AuthenticationError) as context:
            await self.auth_backend.authenticate(self.conn)
        
        self.assertEqual(str(context.exception), "Invalid JWT token")

if __name__ == "__main__":
    unittest.main()



import unittest

from shraga_common.utils import extract_user_org


class TestExtractUserOrg(unittest.TestCase):
    
    def test_extract_user_org(self):
        """Test extract_user_org with different user_id formats"""
        test_cases = [
            ("alice@techcorp.com", "techcorp.com"),
            ("user@gmail.com", ""),
            ("username123", ""),
            ("", ""),
            (None, ""),
        ]
        
        for user_id, expected_org in test_cases:
            with self.subTest(user_id=user_id, expected_org=expected_org):
                self.assertEqual(extract_user_org(user_id), expected_org)

    def test_case_insensitive_common_domains(self):
        """Test that common domains are case insensitive"""
        self.assertEqual(extract_user_org("user@GMAIL.COM"), "")
        self.assertEqual(extract_user_org("user@Gmail.Com"), "")
        self.assertEqual(extract_user_org("user@YAHOO.COM"), "")


if __name__ == '__main__':
    unittest.main()